""" The loopy specific generators """
from __future__ import absolute_import

from dune.perftool.generation import (get_counter,
                                      generator_factory,
                                      no_caching,
                                      preamble,
                                      )
from dune.perftool.generation.context import get_global_context_value
from dune.perftool.error import PerftoolLoopyError

import loopy
import numpy

iname = generator_factory(item_tags=("iname",))
function_mangler = generator_factory(item_tags=("mangler",))
silenced_warning = generator_factory(item_tags=("silenced_warning",), no_deco=True)


class DuneGlobalArg(loopy.GlobalArg):
    allowed_extra_kwargs = loopy.GlobalArg.allowed_extra_kwargs + ["managed"]


@generator_factory(item_tags=("argument", "globalarg"),
                   cache_key_generator=lambda n, **kw: n)
def globalarg(name, shape=loopy.auto, managed=True, **kw):
    if isinstance(shape, str):
        shape = (shape,)
    dtype = kw.pop("dtype", numpy.float64)
    return DuneGlobalArg(name, dtype=dtype, shape=shape, managed=managed, **kw)


@generator_factory(item_tags=("argument", "constantarg"),
                   cache_key_generator=lambda n, **kw: n)
def constantarg(name, shape=loopy.auto, **kw):
    if isinstance(shape, str):
        shape = (shape,)
    dtype = kw.pop("dtype", numpy.float64)
    return loopy.GlobalArg(name, dtype=dtype, shape=shape, **kw)


@generator_factory(item_tags=("argument", "valuearg"),
                   cache_key_generator=lambda n, **kw: n)
def valuearg(name, **kw):
    return loopy.ValueArg(name, **kw)


@generator_factory(item_tags=("domain",))
def domain(iname, shape):
    if isinstance(shape, str):
        valuearg(shape)
    return "{{ [{0}] : 0<={0}<{1} }}".format(iname, shape)


def get_temporary_name():
    return 'expr_{}'.format(str(get_counter('__temporary').zfill(4)))


@generator_factory(item_tags=("temporary",), cache_key_generator=lambda n, **kw: n)
def temporary_variable(name, **kwargs):
    from dune.perftool.loopy.temporary import DuneTemporaryVariable
    return DuneTemporaryVariable(name, scope=loopy.temp_var_scope.PRIVATE, **kwargs)


# Now define generators for instructions. To ease dependency handling of instructions
# these generators are a bit more involved... We apply the following procedure:
# There is one generator that returns the unique id and forwards to a generator that
# actually adds the instruction. Hashing is done based on the code snippet.


@generator_factory(item_tags=("instruction", "cinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['code'],
                   )
def c_instruction_impl(**kw):
    kw.setdefault('assignees', [])
    from pymbolic.primitives import Variable
    kw['assignees'] = frozenset(Variable(i) for i in kw['assignees'])
    inames = kw.pop('inames', kw.get('forced_iname_deps', []))

    return loopy.CInstruction(inames, **kw)


@generator_factory(item_tags=("instruction", "exprinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['expression'],
                   )
def expr_instruction_impl(**kw):
    if 'assignees' in kw:
        from pymbolic.primitives import Variable
        kw['assignees'] = frozenset(Variable(i) for i in kw['assignees'])
    return loopy.ExpressionInstruction(**kw)


@generator_factory(item_tags=("instruction", "callinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['expression'],
                   )
def call_instruction_impl(**kw):
    return loopy.CallInstruction(**kw)


def _insn_cache_key(code=None, expression=None, **kwargs):
    if code is not None:
        return code
    if expression is not None:
        return expression
    raise ValueError("Please specify either code or expression for instruction!")


@generator_factory(item_tags=("insn_id",), cache_key_generator=_insn_cache_key)
def instruction(code=None, expression=None, **kwargs):
    assert (code is not None) or (expression is not None)
    assert not ((code is not None) and (expression is not None))
    assert 'id' not in kwargs

    # Get an ID for this instruction
    id = 'insn_{}'.format(str(get_counter('__insn_id')).zfill(4))

    # Now create the actual instruction
    if code:
        c_instruction_impl(id=id, code=code, **kwargs)
    if expression:
        if 'assignees' in kwargs and len(kwargs['assignees']) == 0:
            call_instruction_impl(id=id, expression=expression, **kwargs)
        else:
            expr_instruction_impl(id=id, expression=expression, **kwargs)

    # return the ID, as it is the only useful information to the user
    return id


@generator_factory(item_tags=("instruction",), cache_key_generator=lambda **kw: kw['id'])
def noop_instruction(**kwargs):
    return loopy.NoOpInstruction(**kwargs)


@generator_factory(item_tags=("transformation",),
                   cache_key_generator=no_caching,
                   )
def transform(trafo, *args):
    return (trafo, args)


@generator_factory(item_tags=("instruction", "barrier"), cache_key_generator=lambda **kw: kw['id'])
def _barrier(**kwargs):
    return loopy.BarrierInstruction(**kwargs)


def barrier(**kwargs):
    assert 'id' not in kwargs
    name = 'barrier_{}'.format(get_counter('barrier'))
    _barrier(id=name, **kwargs)
    return name