import copy

from pymbolic.mapper.substitutor import substitute

from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      barrier,
                                      domain,
                                      function_mangler,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.spaces import name_lfs
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           name_theta,
                                           name_theta_transposed,
                                           )
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from loopy import Reduction, GlobalArg
from loopy.symbolic import FunctionIdentifier

from pytools import product


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


def setup_theta(element, restriction, component, a_matrices, buffer_name):
    number_basis = product(mat.cols for mat in a_matrices)
    shape = (number_basis,)
    inp = get_buffer_temporary(buffer_name,
                               shape=shape)
    silenced_warning('read_no_write({})'.format(inp))

    # Write initial coefficients into buffer
    basisiname = sumfact_iname(number_basis, "basis")
    lfs = name_lfs(element, restriction, component)
    container = name_coefficientcontainer(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),))
    return instruction(assignee=assignee,
                       expression=coeff,
                       )


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension

    # Collect buffers we need
    buffers = []

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.argument.index:
        for i in accterm.argument.index._indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))
                for i in range(dim):
                    buffers.append(name_test_function_contribution(accterm.argument))
    else:
        buffers.append(name_test_function_contribution(accterm.argument))

    # Collect a list of instruction IDs of the contributions
    contribution_ids = []

    # TODO covers only 2D
    for i, buf in enumerate(buffers):
        # Get the a matrices needed for this accumulation term
        theta_transposed = name_theta_transposed()
        rows = basis_functions_per_direction()
        cols = quadrature_points_per_direction()
        theta_matrix = AMatrix(theta_transposed, rows, cols)

        # If this is a gradient we need different matrices
        if accterm.argument.index:
            from dune.perftool.sumfact.amatrix import name_dtheta_transposed
            dtheta_transposed = name_dtheta_transposed()
            rows = basis_functions_per_direction()
            cols = quadrature_points_per_direction()
            dtheta_matrix = AMatrix(dtheta_transposed, rows, cols)

            a_matrices = [theta_matrix] * dim
            a_matrices[i] = dtheta_matrix
            a_matrices = tuple(a_matrices)

        else:
            a_matrices = (theta_matrix, theta_matrix)

        temp = initialize_buffer(buf,
                                 base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                                 num=2
                                 ).get_temporary(shape=(quadrature_points_per_direction(),) * dim,
                                                 dim_tags="f,f")

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        expression = copy.deepcopy(pymbolic_expr)
        for iname in additional_inames:
            replace_dict[Variable(iname)] = i
        expression = substitute(expression, replace_dict)

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = Subscript(Variable(temp), tuple(Variable(i) for i in quadrature_inames()))
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  )
        contribution_ids.append(contrib_dep)

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (step 3)
        result, insn_dep = sum_factorization_kernel(a_matrices,
                                                    buf,
                                                    insn_dep=frozenset({contrib_dep}),
                                                    additional_inames=frozenset(visitor.inames),
                                                    )

        inames = tuple(sumfact_iname(mat.rows, 'accum') for mat in a_matrices)

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
        test_lfs.index = flatten_index(tuple(Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
        rank = 2 if visitor.inames else 1
        if rank == 2:
            ansatz_lfs.index = flatten_index(tuple(Variable(i) for i in visitor.inames),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Construct the expression representing "{r,jac}.accumulate(..)"
        accum = name_accumulation_variable()
        expr = Call(PDELabAccumulationFunction(accum, rank),
                    (ansatz_lfs.get_args() +
                     test_lfs.get_args() +
                     (Subscript(Variable(result), tuple(Variable(i) for i in inames)),)
                     )
                    )

        instruction(assignees=(),
                    expression=expr,
                    forced_iname_deps=frozenset(inames + visitor.inames),
                    forced_iname_deps_is_final=True,
                    depends_on=insn_dep,
                    )

        # Mark the transformation that moves the quadrature loop inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)

    # Maybe try to vectorize!
    if get_option("vectorize"):
        from dune.perftool.loopy.transformations.collect_rotate import collect_vector_data_rotate
        transform(collect_vector_data_rotate, contribution_ids, quadrature_inames())


def sum_factorization_kernel(a_matrices, buf, insn_dep=frozenset({}), additional_inames=frozenset({})):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buf: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input.
    insn_dep: an instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    """
    insn_dep = frozenset({barrier(depends_on=insn_dep,
                                  within_inames=additional_inames,
                                  )})

    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.cols, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        inp = get_buffer_temporary(buf,
                                   shape=inp_shape,
                                   dim_tags="f,f")

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buf)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.rows, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))

        out = get_buffer_temporary(buf,
                                   shape=out_shape,
                                   dim_tags="c,c")

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")
        k = sumfact_iname(a_matrix.cols, "red")

        # Construct the matrix-matrix-multiplication expression a_ik*in_kj
        from dune.perftool.sumfact.amatrix import ColMajorAccess
        prod = Product((Call(ColMajorAccess(a_matrix.a_matrix), (Variable(i), Variable(k))),
                        Subscript(Variable(inp), (Variable(k), Variable(j)))
                        ))

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=Subscript(Variable(out), (Variable(i), Variable(j))),
                                          expression=Reduction("sum", k, prod),
                                          forced_iname_deps=frozenset({i, j}).union(additional_inames),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep,
                                          )
                              })

    out = get_buffer_temporary(buf,
                               shape=out_shape,
                               dim_tags="f,f")
    silenced_warning('read_no_write({})'.format(out))

    return out, insn_dep
